import sysconfig

# register preprocessors
import orangecontrib.snom.widgets.preprocessors  # noqa: F401

# Category metadata.

# Category icon show in the menu
ICON = 'icons/category.svg'

# Background color for category background in menu
# and widget icon background in workflow.
BACKGROUND = "light-blue"

# Location of widget help files.
WIDGET_HELP_PATH = (
    # Development documentation
    # You need to build help pages manually using
    # 'make htmlhelp' inside doc folder
    ('{DEVELOP_ROOT}/doc/_build/htmlhelp/index.html', None),
    # Documentation included in wheel
    # Correct DATA_FILES entry is needed in setup.py and documentation has to be built
    # before the wheel is created.
    (
        '{}/help/orange-snom/index.html'.format(sysconfig.get_path('data')),
        None,
    ),
    # Online documentation url, used when the local documentation is not available.
    # Url should point to a page with a section Widgets. This section should
    # includes links to documentation pages of each widget. Matching is
    # performed by comparing link caption to widget name.
    ('http://orange-snom.readthedocs.io/en/latest/', ''),
)



# this is just an example of registration

from AnyQt.QtWidgets import QFormLayout

from Orange.data import Domain
from Orange.preprocess import Preprocess
from orangecontrib.snom.widgets.preprocessors.registry import preprocess_image_editors

from orangecontrib.spectroscopy.preprocess import SelectColumn, CommonDomain

from orangecontrib.spectroscopy.widgets.preprocessors.utils import BaseEditorOrange
from orangecontrib.spectroscopy.widgets.gui import lineEditFloatRange

from pySNOM.images import RotatePhase


class AddFeature(SelectColumn):
    InheritEq = True


class _PhaseRotCommon(CommonDomain):
    def __init__(self, degree, domain):
        super().__init__(domain)
        self.degree = degree

    def transformed(self, data):
        return RotatePhase(degree=self.degree).transform(data.X)


class PhaseRotation(Preprocess):
    def __init__(self, degree=0.0):
        self.degree = degree

    def __call__(self, data):
        common = _PhaseRotCommon(self.degree, data.domain)
        atts = [
            a.copy(compute_value=AddFeature(i, common))
            for i, a in enumerate(data.domain.attributes)
        ]
        domain = Domain(atts, data.domain.class_vars, data.domain.metas)
        return data.from_table(domain, data)


class PhaseRotationEditor(BaseEditorOrange):
    name = "Rotate phase"
    qualname = "orangecontrib.snom.phase_rotation_test"

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.degree = 0.0

        form = QFormLayout()
        degreeedit = lineEditFloatRange(self, self, "degree", callback=self.edited.emit)
        # degreeedit = slideEditFloatRange(self, self, "degree", callback=self.edited.emit)
        form.addRow("degree", degreeedit)
        self.controlArea.setLayout(form)

    def activateOptions(self):
        pass  # actions when user starts changing options

    def setParameters(self, params):
        self.degree = params.get("degree", 0.0)

    @classmethod
    def createinstance(cls, params):
        params = dict(params)
        degree = float(params.get("degree", 0.0))
        return PhaseRotation(degree=degree)

    def set_preview_data(self, data):
        if data:
            pass  # TODO any settings


preprocess_image_editors.register(PhaseRotationEditor, 300)



from AnyQt.QtWidgets import QFormLayout, QLabel

from orangecontrib.spectroscopy.preprocess import MissingReferenceException
from orangecontrib.spectroscopy.widgets.preprocessors.utils import (
    BaseEditorOrange,
    REFERENCE_DATA_PARAM,
)

from pySNOM.images import SelfReference, DataTypes

from orangecontrib.snom.preprocess.utils import (
    PreprocessImageOpts2DOnlyWholeReference,
)
from orangecontrib.snom.widgets.preprocessors.registry import preprocess_image_editors


class SelfRef(PreprocessImageOpts2DOnlyWholeReference):
    def __init__(self, reference):
        self.reference = reference
        if self.reference is None:
            raise MissingReferenceException("Self-referencing needs a reference")

    def transform_image(self, image, ref_image, data):
        datatype = data.attributes.get("measurement.signaltype", "Phase")
        return SelfReference(
            referencedata=ref_image, datatype=DataTypes[datatype]
        ).transform(image)


class SelfRefEditor(BaseEditorOrange):
    name = "Self-referencing"
    qualname = "orangecontrib.snom.self_reference"

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.reference = None

        form = QFormLayout()
        self.reference_info = QLabel("Reference data from input!")
        form.addRow(self.reference_info)
        self.controlArea.setLayout(form)

    def activateOptions(self):
        pass  # actions when user starts changing options

    def setParameters(self, params):
        self.update_reference_info()

    def set_reference_data(self, reference):
        self.reference = reference
        self.update_reference_info()

    def update_reference_info(self):
        if not self.reference:
            self.reference_info.setText("Reference: missing!")
            self.reference_info.setStyleSheet("color: red")
        else:
            rinfo = "Reference order: N"
            self.reference_info.setText(rinfo)
            self.reference_info.setStyleSheet("color: black")

    @classmethod
    def createinstance(cls, params):
        params = dict(params)
        reference = params.get(REFERENCE_DATA_PARAM, None)
        return SelfRef(reference=reference)

    def set_preview_data(self, data):
        if data:
            pass  # TODO any settings


preprocess_image_editors.register(SelfRefEditor, 400)



from AnyQt.QtWidgets import QFormLayout

from orangecontrib.spectroscopy.widgets.preprocessors.utils import BaseEditorOrange
from orangecontrib.spectroscopy.widgets.gui import lineEditIntRange

from pySNOM.images import BackgroundPolyFit, DataTypes

from orangecontrib.snom.preprocess.utils import (
    PreprocessImageOpts2DOnlyWhole,
)
from orangecontrib.snom.widgets.preprocessors.registry import preprocess_image_editors


class BackGroundFit(PreprocessImageOpts2DOnlyWhole):
    def __init__(self, xorder=1, yorder=1):
        self.xorder = xorder
        self.yorder = yorder

    def transform_image(self, image, data):
        datatype = data.attributes.get("measurement.signaltype", "Phase")
        d, b = BackgroundPolyFit(
            xorder=self.xorder, yorder=self.yorder, datatype=DataTypes[datatype]
        ).transform(image)
        return d


class BackGroundFitEditor(BaseEditorOrange):
    name = "Polynomial background fit"
    qualname = "orangecontrib.snom.background_fit_test"

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.xorder = 1
        self.yorder = 1

        form = QFormLayout()
        xorderedit = lineEditIntRange(self, self, "xorder", callback=self.edited.emit)
        yorderedit = lineEditIntRange(self, self, "yorder", callback=self.edited.emit)
        form.addRow("xorder", xorderedit)
        form.addRow("yorder", yorderedit)
        self.controlArea.setLayout(form)

    def activateOptions(self):
        pass  # actions when user starts changing options

    def setParameters(self, params):
        self.xorder = params.get("xorder", 1)
        self.yorder = params.get("yorder", 1)

    @classmethod
    def createinstance(cls, params):
        params = dict(params)
        xorder = float(params.get("xorder", 1))
        yorder = float(params.get("yorder", 1))
        return BackGroundFit(xorder=xorder, yorder=yorder)

    def set_preview_data(self, data):
        if data:
            pass  # TODO any settings


preprocess_image_editors.register(BackGroundFitEditor, 400)



import numpy as np
from AnyQt.QtWidgets import QFormLayout

from orangewidget.gui import comboBox

from pySNOM.images import LineLevel, DataTypes

from orangecontrib.spectroscopy.widgets.preprocessors.utils import BaseEditorOrange

from orangecontrib.snom.widgets.preprocessors.registry import preprocess_image_editors
from orangecontrib.snom.preprocess.utils import (
    PreprocessImageOpts2DOnlyWhole,
)


class LineLevelProcessor(PreprocessImageOpts2DOnlyWhole):
    def __init__(self, method="median"):
        self.method = method

    def transform_image(self, image, data):
        datatype = data.attributes.get("measurement.signaltype", "Phase")
        processed = LineLevel(
            method=self.method, datatype=DataTypes[datatype]
        ).transform(image)
        if self.method == 'difference':
            # add a row of NaN, so that the size matches
            processed = np.vstack((processed, np.full((1, image.shape[1]), np.nan)))
        return processed


class LineLevelEditor(BaseEditorOrange):
    name = "Line leveling"
    qualname = "orangecontrib.snom.line_level_test"

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.method = 'median'

        form = QFormLayout()
        self.levelmethod_cb = comboBox(self, self, "method", callback=self.setmethod)
        self.levelmethod_cb.addItems(['median', 'mean', 'difference'])
        form.addRow("Leveling method", self.levelmethod_cb)
        self.controlArea.setLayout(form)

    def activateOptions(self):
        pass  # actions when user starts changing options

    def setmethod(self):
        self.method = self.levelmethod_cb.currentText()
        self.edited.emit()

    def setParameters(self, params):
        self.method = params.get("method", "median")

    @classmethod
    def createinstance(cls, params):
        params = dict(params)
        method = params.get("method", "median")
        return LineLevelProcessor(method=method)

    def set_preview_data(self, data):
        if data:
            pass  # TODO any settings


preprocess_image_editors.register(LineLevelEditor, 200)



import numpy as np

from Orange.data import Domain
from Orange.preprocess import Preprocess
from orangecontrib.spectroscopy.preprocess import (
    CommonDomain,
    SelectColumn,
    WrongReferenceException,
)
from orangecontrib.spectroscopy.utils import (
    InvalidAxisException,
    values_to_linspace,
    index_values,
)


class PreprocessImageOpts(Preprocess):
    pass


class PreprocessImageOpts2D(PreprocessImageOpts):
    def __call__(self, data, image_opts):
        common = self.image_transformer(data, image_opts)
        at = data.domain[image_opts["attr_value"]].copy(
            compute_value=SelectColumn(0, common)
        )
        domain = domain_with_single_attribute_in_x(at, data.domain)
        return data.transform(domain)

    def image_transformer(self, data, image_opts):
        raise NotImplementedError


class NoComputeValue:
    def __call__(self, data):
        return np.full(len(data), np.nan)


def _prepare_domain_for_image(data, image_opts):
    at = data.domain[image_opts["attr_value"]].copy(compute_value=NoComputeValue())
    return domain_with_single_attribute_in_x(at, data.domain)


def _prepare_table_for_image(data, image_opts):
    odata = data
    domain = _prepare_domain_for_image(data, image_opts)
    data = data.transform(domain)
    if len(data):
        with data.unlocked(data.X):
            data.X[:, 0] = odata.get_column(image_opts["attr_value"], copy=True)
    return data


def _image_from_table(data, image_opts):
    hypercube, _, indices = get_ndim_hyperspec(
        data, (image_opts["attr_y"], image_opts["attr_x"])
    )
    return hypercube[:, :, 0], indices


class PreprocessImageOpts2DOnlyWhole(PreprocessImageOpts):
    def __call__(self, data, image_opts):
        data = _prepare_table_for_image(data, image_opts)
        try:
            image, indices = _image_from_table(data, image_opts)
            transformed = self.transform_image(image, data)
            col = transformed[indices].reshape(-1)
        except InvalidAxisException:
            col = np.full(len(data), np.nan)
        if len(data):
            with data.unlocked(data.X):
                data.X[:, 0] = col
        return data

    def transform_image(self, image, data):
        """
        image: a numpy 2D array where image[y,x] is the value in image row y and column x
        data: image data set (used for passing meta data)
        """
        raise NotImplementedError


class PreprocessImageOpts2DOnlyWholeReference(PreprocessImageOpts):
    def __call__(self, data, image_opts):
        data = _prepare_table_for_image(data, image_opts)
        reference = _prepare_table_for_image(self.reference, image_opts)
        try:
            image, indices = _image_from_table(data, image_opts)
            ref_image, _ = _image_from_table(reference, image_opts)
            if image.shape != ref_image.shape:
                raise WrongReferenceException("Reference data should have length 1")
            transformed = self.transform_image(image, ref_image, data)
            col = transformed[indices].reshape(-1)
        except InvalidAxisException:
            col = np.full(len(data), np.nan)
        if len(data):
            with data.unlocked(data.X):
                data.X[:, 0] = col
        return data

    def transform_image(self, image, ref_image, data):
        """
        image: a numpy 2D array where image[y,x] is the value in image row y and column x
        ref_image: a numpy 2D array where image[y,x] is the value in image row y and column x
        data: image data set (used for passing meta data)
        """
        raise NotImplementedError


def axes_to_ndim_linspace(coordinates):
    # modified to avoid domains as much as possible
    ls = []
    indices = []

    for i in range(coordinates.shape[1]):
        coor = coordinates[:, i]
        lsa = values_to_linspace(coor)
        if lsa is None:
            raise InvalidAxisException(i)
        ls.append(lsa)
        indices.append(index_values(coor, lsa))

    return ls, tuple(indices)


def get_ndim_hyperspec(data, attrs):
    # mostly copied from orangecontrib.spectroscopy.utils,
    # but returns the indices too
    # also avoid Orange domains as much as possible
    coordinates = np.hstack([data.get_column(a).reshape(-1, 1) for a in attrs])

    ls, indices = axes_to_ndim_linspace(coordinates)

    # set data
    new_shape = tuple([lsa[2] for lsa in ls]) + (data.X.shape[1],)
    hyperspec = np.ones(new_shape) * np.nan

    hyperspec[indices] = data.X

    return hyperspec, ls, indices


def domain_with_single_attribute_in_x(attribute, domain):
    """Create a domain with only the attribute in domain.attributes and ensure
    that the same attribute is removed from metas and class_vars if it was present
    there."""
    class_vars = [a for a in domain.class_vars if a.name != attribute.name]
    metas = [a for a in domain.metas if a.name != attribute.name]
    return Domain([attribute], class_vars, metas)


class CommonDomainImage2D(CommonDomain):
    def __init__(self, domain: Domain, image_opts: dict):
        self.domain = domain
        self.image_opts = image_opts

    def __call__(self, data):
        data = self.transform_domain(data)
        vat = data.domain[self.image_opts["attr_value"]]
        ndom = domain_with_single_attribute_in_x(vat, data.domain)
        data = data.transform(ndom)
        try:
            hypercube, _, indices = get_ndim_hyperspec(
                data, (self.image_opts["attr_y"], self.image_opts["attr_x"])
            )
            image = hypercube[:, :, 0]
            transformed = self.transform_image(image)
            return transformed[indices].reshape(-1, 1)
        except InvalidAxisException:
            return np.full((len(data), 1), np.nan)
        return self.transformed(data)

    def transform_image(self, image):
        """
        image: a numpy 2D array where image[y,x] is the value in image row y and column x
        """
        raise NotImplementedError



import numpy as np

from orangewidget.tests.utils import excepthook_catch

from Orange.data import Table, Domain
from Orange.widgets.tests.base import WidgetTest
from Orange.preprocess.preprocess import Preprocess

from orangecontrib.spectroscopy.tests import spectral_preprocess
from orangecontrib.spectroscopy.tests.spectral_preprocess import (
    pack_editor,
    wait_for_preview,
)
from orangecontrib.spectroscopy.widgets.preprocessors.misc import (
    CutEditor,
    SavitzkyGolayFilteringEditor,
)

from orangecontrib.snom.widgets.owpreprocessimage import OWPreprocessImage
from orangecontrib.snom.widgets.preprocessors.registry import preprocess_image_editors


PREPROCESSORS = list(map(pack_editor, preprocess_image_editors.sorted()))


WHITELIGHT = Table("whitelight.gsf")


class TestAllPreprocessors(WidgetTest):
    def test_allpreproc_indv(self):
        data = WHITELIGHT
        for p in PREPROCESSORS:
            with self.subTest(p.viewclass):
                self.widget = self.create_widget(OWPreprocessImage)
                self.send_signal("Data", data)
                self.widget.add_preprocessor(p)
                self.widget.commit.now()
                wait_for_preview(self.widget)
                self.wait_until_finished(timeout=10000)

    def test_allpreproc_indv_empty(self):
        data = WHITELIGHT
        for p in PREPROCESSORS:
            with self.subTest(p.viewclass):
                self.widget = self.create_widget(OWPreprocessImage)
                self.send_signal("Data", data[:0])
                self.widget.add_preprocessor(p)
                self.widget.commit.now()
                wait_for_preview(self.widget)
                self.wait_until_finished(timeout=10000)
        # no attributes
        data = data.transform(
            Domain([], class_vars=data.domain.class_vars, metas=data.domain.metas)
        )
        for p in PREPROCESSORS:
            with self.subTest(p.viewclass, type="no attributes"):
                self.widget = self.create_widget(OWPreprocessImage)
                self.send_signal("Data", data)
                self.widget.add_preprocessor(p)
                self.widget.commit.now()
                wait_for_preview(self.widget)
                self.wait_until_finished(timeout=10000)


class TestOWPreprocess(WidgetTest):
    def setUp(self):
        self.widget = self.create_widget(OWPreprocessImage)

    def test_load_unload(self):
        self.send_signal("Data", Table("iris.tab"))
        self.send_signal("Data", None)

    def failing_saving_preprocessors(self):
        settings = self.widget.settingsHandler.pack_data(self.widget)
        self.assertEqual([], settings["storedsettings"]["preprocessors"])
        self.widget.add_preprocessor(self.widget.PREPROCESSORS[0])
        settings = self.widget.settingsHandler.pack_data(self.widget)
        self.assertEqual(
            self.widget.PREPROCESSORS[0].qualname,
            settings["storedsettings"]["preprocessors"][0][0],
        )

    def test_output_preprocessor_without_data(self):
        self.widget.add_preprocessor(pack_editor(CutEditor))
        self.widget.commit.now()
        self.wait_until_finished()
        out = self.get_output(self.widget.Outputs.preprocessor)
        self.assertIsInstance(out, Preprocess)

    def test_empty_no_inputs(self):
        self.widget.commit.now()
        self.wait_until_finished()
        p = self.get_output(self.widget.Outputs.preprocessor)
        d = self.get_output(self.widget.Outputs.preprocessed_data)
        self.assertEqual(None, p)
        self.assertEqual(None, d)

    def test_no_preprocessors(self):
        data = WHITELIGHT
        self.send_signal(self.widget.Inputs.data, data)
        self.widget.commit.now()
        self.wait_until_finished()
        d = self.get_output(self.widget.Outputs.preprocessed_data)
        self.assertEqual(WHITELIGHT, d)

    def test_widget_vs_manual(self):
        data = WHITELIGHT
        self.send_signal(self.widget.Inputs.data, data)
        self.widget.add_preprocessor(pack_editor(CutEditor))
        self.widget.add_preprocessor(pack_editor(SavitzkyGolayFilteringEditor))
        self.widget.commit.now()
        self.wait_until_finished()
        p = self.get_output(self.widget.Outputs.preprocessor)
        d = self.get_output(self.widget.Outputs.preprocessed_data)
        manual = p(data)
        np.testing.assert_equal(d.X, manual.X)

    def test_invalid_preprocessors(self):
        settings = {"storedsettings": {"preprocessors": [("xyz.abc.notme", {})]}}
        with self.assertRaises(KeyError):
            with excepthook_catch(raise_on_exit=True):
                widget = self.create_widget(OWPreprocessImage, settings)
                self.assertTrue(widget.Error.loading.is_shown())


class TestPreprocessWarning(spectral_preprocess.TestWarning):
    widget_cls = OWPreprocessImage

    def test_exception_preview_after_data(self):
        self.editor.raise_exception = True
        self.editor.edited.emit()
        wait_for_preview(self.widget)
        self.assertIsNone(self.widget.curveplot_after.data)

        self.editor.raise_exception = False
        self.editor.edited.emit()
        wait_for_preview(self.widget)
        self.assertIsNotNone(self.widget.curveplot_after.data)



import numpy as np

import Orange
import Orange.data
from Orange.preprocess.preprocess import Preprocess

from orangecontrib.spectroscopy.preprocess.utils import (
    SelectColumn,
    CommonDomainOrderUnknowns,
)


class _PhaseUnwrapCommon(CommonDomainOrderUnknowns):
    # A possible optimization in the future:
    # the "order" functionality of CommonDomainOrderUnknowns is not needed,
    # but does not break anything either

    def transformed(self, X, _):  # noqa: N803
        return np.unwrap(X)


class PhaseUnwrap(Preprocess):
    """
    Unwrap phase values using numpy.unwrap defaults or bypass the data.
    """

    def __call__(self, data):
        common = _PhaseUnwrapCommon(data.domain)
        atts = [
            a.copy(compute_value=SelectColumn(i, common))
            for i, a in enumerate(data.domain.attributes)
        ]
        domain = Orange.data.Domain(atts, data.domain.class_vars, data.domain.metas)
        return data.from_table(domain, data)



import time

from AnyQt.QtCore import Qt

from Orange.data import Domain, DiscreteVariable, ContinuousVariable
from Orange.widgets.settings import DomainContextHandler
from Orange.widgets.utils.itemmodels import DomainModel
from orangecontrib.snom.widgets.preprocessors.registry import preprocess_image_editors
from orangecontrib.snom.preprocess.utils import PreprocessImageOpts
from orangewidget import gui
from orangewidget.settings import SettingProvider, ContextSetting, Setting

import Orange.data
from Orange import preprocess
from Orange.widgets.widget import Output, Input

from orangecontrib.spectroscopy.widgets.owhyper import BasicImagePlot

from orangecontrib.spectroscopy.widgets.owpreprocess import (
    GeneralPreprocess,
    create_preprocessor,
    InterruptException,
    PreviewRunner,
)

from orangewidget.widget import Msg


class AImagePlot(BasicImagePlot):
    attr_x = None  # not settings, set from the parent class
    attr_y = None

    def __init__(self, parent):
        super().__init__(parent)
        self.axes_settings_box.hide()
        self.rgb_settings_box.hide()

    def add_selection_actions(self, _):
        pass

    def clear_markings(self):
        pass


class ImagePreviews:
    curveplot = SettingProvider(AImagePlot)
    curveplot_after = SettingProvider(AImagePlot)

    value_type = 1

    def __init__(self):
        # the name of curveplot is kept because GeneralPreprocess
        # expects these names
        self.curveplot = AImagePlot(self)
        self.curveplot_after = AImagePlot(self)

    def shutdown(self):
        self.curveplot.shutdown()
        self.curveplot_after.shutdown()


class SpectralImagePreprocess(GeneralPreprocess, ImagePreviews, openclass=True):
    def __init__(self):
        ImagePreviews.__init__(self)
        super().__init__()

    def onDeleteWidget(self):
        super().onDeleteWidget()
        ImagePreviews.shutdown(self)


def execute_with_image_opts(pp, data, image_opts):
    if isinstance(pp, PreprocessImageOpts):
        return pp(data, image_opts)
    return pp(data)


class ImagePreviewRunner(PreviewRunner):
    def show_preview(self, show_info_anyway=False):
        """Shows preview and also passes preview data to the widgets"""
        master = self.master
        self.preview_pos = master.flow_view.preview_n()
        self.last_partial = None
        self.show_info_anyway = show_info_anyway
        self.preview_data = None
        self.after_data = None
        pp_def = [
            master.preprocessormodel.item(i)
            for i in range(master.preprocessormodel.rowCount())
        ]
        if master.data is not None:
            data = master.sample_data(master.data)
            image_opts = master.image_opts()
            self.start(
                self.run_preview,
                data,
                master.reference_data,
                image_opts,
                pp_def,
                master.process_reference,
            )
        else:
            master.curveplot.set_data(None)
            master.curveplot_after.set_data(None)

    @staticmethod
    def run_preview(
        data: Orange.data.Table,
        reference: Orange.data.Table,
        image_opts,
        pp_def,
        process_reference,
        state,
    ):
        def progress_interrupt(i: float):
            if state.is_interruption_requested():
                raise InterruptException

        n = len(pp_def)
        orig_data = data
        for i in range(n):
            progress_interrupt(0)
            state.set_partial_result((i, data, reference))
            item = pp_def[i]
            pp = create_preprocessor(item, reference)
            data = execute_with_image_opts(pp, data, image_opts)
            progress_interrupt(0)
            if process_reference and reference is not None and i != n - 1:
                reference = execute_with_image_opts(pp, reference, image_opts)
        progress_interrupt(0)
        state.set_partial_result((n, data, None))
        return orig_data, data


class SpectralImagePreprocessReference(SpectralImagePreprocess, openclass=True):
    class Inputs(SpectralImagePreprocess.Inputs):
        reference = Input("Reference", Orange.data.Table)

    @Inputs.reference
    def set_reference(self, reference):
        self.reference_data = reference


class OWPreprocessImage(SpectralImagePreprocessReference):
    name = "Preprocess image"
    id = "orangecontrib.snom.widgets.preprocessimage"
    description = "Process image"
    icon = "icons/preprocessimage.svg"
    priority = 1010

    settings_version = 2

    settingsHandler = DomainContextHandler()

    _max_preview_spectra = 1000000
    preview_curves = Setting(100000)

    editor_registry = preprocess_image_editors
    BUTTON_ADD_LABEL = "Add preprocessor..."

    attr_value = ContextSetting(None)
    attr_x = ContextSetting(None, exclude_attributes=True)
    attr_y = ContextSetting(None, exclude_attributes=True)

    class Outputs:
        preprocessed_data = Output("Integrated Data", Orange.data.Table, default=True)
        preprocessor = Output("Preprocessor", preprocess.preprocess.Preprocess)

    class Warning(SpectralImagePreprocess.Warning):
        threshold_error = Msg("Low slider should be less than High")

    class Error(SpectralImagePreprocess.Error):
        image_too_big = Msg("Image for chosen features is too big ({} x {}).")

    class Information(SpectralImagePreprocess.Information):
        not_shown = Msg("Undefined positions: {} data point(s) are not shown.")

    def image_values(self):
        attr_value = self.attr_value.name if self.attr_value else None
        return lambda data, attr=attr_value: data.transform(Domain([data.domain[attr]]))

    def image_values_fixed_levels(self):
        return None

    def __init__(self):
        self.markings_list = []
        super().__init__()

        self.preview_runner = ImagePreviewRunner(self)

        self.feature_value_model = DomainModel(
            order=(
                DomainModel.ATTRIBUTES,
                DomainModel.Separator,
                DomainModel.CLASSES,
                DomainModel.Separator,
                DomainModel.METAS,
            ),
            valid_types=ContinuousVariable,
        )

        common_options = {
            "labelWidth": 50,
            "orientation": Qt.Horizontal,
            "sendSelectedValue": True,
        }

        self.feature_value = gui.comboBox(
            self.preview_settings_box,
            self,
            "attr_value",
            label="Show",
            contentsLength=12,
            searchable=True,
            callback=self.update_feature_value,
            model=self.feature_value_model,
            **common_options
        )

        self.xy_model = DomainModel(
            DomainModel.METAS | DomainModel.CLASSES, valid_types=DomainModel.PRIMITIVE
        )

        self.cb_attr_x = gui.comboBox(
            self.preview_settings_box,
            self,
            "attr_x",
            label="Axis x",
            callback=self.update_attr,
            model=self.xy_model,
            **common_options
        )
        self.cb_attr_y = gui.comboBox(
            self.preview_settings_box,
            self,
            "attr_y",
            label="Axis y",
            callback=self.update_attr,
            model=self.xy_model,
            **common_options
        )

        self.contextAboutToBeOpened.connect(lambda x: self.init_interface_data(x[0]))

        self.preview_runner.preview_updated.connect(self.redraw_data)

    def update_attr(self):
        self.curveplot.attr_x = self.attr_x
        self.curveplot.attr_y = self.attr_y
        self.curveplot_after.attr_x = self.attr_x
        self.curveplot_after.attr_y = self.attr_y
        self.on_modelchanged()

    def update_feature_value(self):
        self.on_modelchanged()

    def redraw_data(self):
        self.curveplot.update_view()
        self.curveplot_after.update_view()

    def init_interface_data(self, data):
        self.init_attr_values(data)
        self.curveplot.init_interface_data(data)
        self.curveplot_after.init_interface_data(data)

    def init_attr_values(self, data):
        domain = data.domain if data is not None else None
        self.feature_value_model.set_domain(domain)
        self.attr_value = (
            self.feature_value_model[0] if self.feature_value_model else None
        )
        self.xy_model.set_domain(domain)
        self.attr_x = self.xy_model[0] if self.xy_model else None
        self.attr_y = self.xy_model[1] if len(self.xy_model) >= 2 else self.attr_x

    def image_opts(self):
        return {
            'attr_x': str(self.attr_x),
            'attr_y': str(self.attr_y),
            'attr_value': str(self.attr_value),
        }

    def create_outputs(self):
        self._reference_compat_warning()
        pp_def = [
            self.preprocessormodel.item(i)
            for i in range(self.preprocessormodel.rowCount())
        ]
        image_opts = self.image_opts()
        self.start(
            self.run_task,
            self.data,
            self.reference_data,
            image_opts,
            pp_def,
            self.process_reference,
        )

    def set_data(self, data):
        self.curveplot.set_data(None)
        self.curveplot_after.set_data(None)

        super().set_data(data)

        self.closeContext()

        def valid_context(data):
            if data is None:
                return False
            annotation_features = [
                v
                for v in data.domain.metas + data.domain.class_vars
                if isinstance(v, (DiscreteVariable, ContinuousVariable))
            ]
            return len(annotation_features) >= 1

        if valid_context(data):
            self.openContext(data)
        else:
            # to generate valid interface even if context was not loaded
            self.contextAboutToBeOpened.emit([data])

        self.update_attr()  # update imageplots attributes from the master

    @staticmethod
    def run_task(
        data: Orange.data.Table,
        reference: Orange.data.Table,
        image_opts,
        pp_def,
        process_reference,
        state,
    ):
        def progress_interrupt(i: float):
            state.set_progress_value(i)
            if state.is_interruption_requested():
                raise InterruptException

        # Protects against running the task in succession many times, as would
        # happen when adding a preprocessor (there, commit() is called twice).
        # Wait 100 ms before processing - if a new task is started in meanwhile,
        # allow that is easily` cancelled.
        for _ in range(10):
            time.sleep(0.010)
            progress_interrupt(0)

        n = len(pp_def)
        plist = []
        for i in range(n):
            progress_interrupt(i / n * 100)
            item = pp_def[i]
            pp = create_preprocessor(item, reference)
            plist.append(pp)
            if data is not None:
                data = execute_with_image_opts(pp, data, image_opts)
            progress_interrupt((i / n + 0.5 / n) * 100)
            if process_reference and reference is not None and i != n - 1:
                reference = execute_with_image_opts(pp, reference, image_opts)
        # if there are no preprocessors, return None instead of an empty list
        preprocessor = preprocess.preprocess.PreprocessorList(plist) if plist else None
        return data, preprocessor


if __name__ == "__main__":  # pragma: no cover
    from Orange.widgets.utils.widgetpreview import WidgetPreview

    WidgetPreview(OWPreprocessImage).run(Orange.data.Table("whitelight.gsf"))



import unittest

import numpy as np

from Orange.data import Table

from orangecontrib.spectroscopy.tests.test_preprocess import (
    TestCommonIndpSamplesMixin,
    SMALLER_COLLAGEN,
)

from orangecontrib.snom.preprocess import PhaseUnwrap


class TestPhaseUnwrap(unittest.TestCase, TestCommonIndpSamplesMixin):
    preprocessors = [PhaseUnwrap()]
    data = SMALLER_COLLAGEN

    def test_simple(self):
        data = Table.from_numpy(None, [[1, 1 + 2 * np.pi]])
        f = PhaseUnwrap()
        fdata = f(data)
        # check that unwrap removes jumps greater that 2*pi
        np.testing.assert_array_equal(fdata, [[1, 1]])



# load and register editors
from orangecontrib.snom.widgets.preprocessors import phase_unwrap  # noqa: F401
from orangecontrib.snom.widgets.preprocessors import background_fit  # noqa: F401
from orangecontrib.snom.widgets.preprocessors import phase_rotation  # noqa: F401
from orangecontrib.snom.widgets.preprocessors import simple_normalize  # noqa: F401
from orangecontrib.snom.widgets.preprocessors import linelevel  # noqa: F401
from orangecontrib.snom.widgets.preprocessors import self_reference  # noqa: F401



from AnyQt.QtWidgets import QVBoxLayout

from orangecontrib.spectroscopy.widgets.preprocessors.utils import BaseEditorOrange
from orangecontrib.spectroscopy.widgets.preprocessors.registry import preprocess_editors
from orangecontrib.snom.preprocess.phase_unwrap import PhaseUnwrap


class PhaseUnwrapEditor(BaseEditorOrange):
    """
    Phase Unwrap editor.
    """

    name = "Phase Unwrap"
    qualname = "orangecontrib.snom.phaseunwrap"

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.controlArea.setLayout(QVBoxLayout())

    def setParameters(self, params):
        pass

    @classmethod
    def createinstance(cls, params):
        return PhaseUnwrap()


preprocess_editors.register(PhaseUnwrapEditor, 1000)



from AnyQt.QtWidgets import QFormLayout

from orangecontrib.spectroscopy.widgets.preprocessors.utils import BaseEditorOrange
from orangecontrib.spectroscopy.widgets.gui import lineEditFloatRange
from orangewidget.gui import comboBox

from pySNOM.images import SimpleNormalize, DataTypes

from orangecontrib.snom.widgets.preprocessors.registry import preprocess_image_editors
from orangecontrib.snom.preprocess.utils import (
    PreprocessImageOpts2DOnlyWhole,
)


class SimpleNorm(PreprocessImageOpts2DOnlyWhole):
    def __init__(self, method, value):
        self.method = method
        self.value = value

    def transform_image(self, image, data):
        datatype = data.attributes.get("measurement.signaltype", "Phase")
        return SimpleNormalize(
            method=self.method, value=self.value, datatype=DataTypes[datatype]
        ).transform(image)


class SimpleNormEditor(BaseEditorOrange):
    name = "Simple normalization"
    qualname = "orangecontrib.snom.simple_normalize"

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent, **kwargs)

        self.method = "manual"
        self.value = 1.0

        form = QFormLayout()
        self.valueedit = lineEditFloatRange(
            self, self, "value", callback=self.edited.emit
        )
        self.cb_method = comboBox(self, self, "method", callback=self.setmethod)
        self.cb_method.addItems(['median', 'mean', 'manual'])
        self.cb_method.setCurrentText('manual')
        form.addRow("method", self.cb_method)
        form.addRow("value", self.valueedit)
        self.controlArea.setLayout(form)

    def setmethod(self):
        if self.cb_method.currentText() != "manual":
            self.valueedit.setEnabled(False)
        else:
            self.valueedit.setEnabled(True)

        self.method = self.cb_method.currentText()
        self.edited.emit()

    def activateOptions(self):
        pass  # actions when user starts changing options

    def setParameters(self, params):
        self.method = params.get("method", "manual")
        self.value = params.get("value", 1)

    @classmethod
    def createinstance(cls, params):
        params = dict(params)
        method = str(params.get("method", "manual"))
        value = float(params.get("value", 1))
        return SimpleNorm(method=method, value=value)

    def set_preview_data(self, data):
        if data:
            pass  # TODO any settings


preprocess_image_editors.register(SimpleNormEditor, 500)
